//=============================================================================
//
//	Copyright (C) 1995,1996,1997,1998 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//
//=============================================================================
//-----------------------------------------------------------------------------
// MiscTableViewCursor.M
//
//		Keyboard cursor methods for MiscTableView.
//
// FIXME: The PostScript pattern used to draw the dotted "keyboard focus" 
//		rectangle gets "out-of-phase" as this view scrolls.	 Consequently it 
//		does not tile correctly.  The Adobe Purple Book section 10.4.4 
//		implies that the pattern has to be recreated each time the view 
//		changes.  (I forsee using -initGState for this.) 
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscTableViewCursor.M,v 1.2 98/03/24 05:41:15 sunshine Exp $
// $Log:		MiscTableViewCursor.M,v $
//	Revision 1.2  98/03/24	05:41:15  sunshine
//	v138: Fixed #import ("..." vs. <...>) for MiscKit compatibility.
//	
//	Revision 1.1  97/11/14	20:44:16  zarnuk
//	v130: Keyboard cursor methods.
//-----------------------------------------------------------------------------
#import <misckit/MiscTableScroll.h>
#import "MiscTableViewPrivate.h"
#import "MiscAssert.h"
#import "MiscTableBorder.h"
#import "MiscMouseTracker.h"

extern "Objective-C" {
#import <appkit/FontManager.h>
}

extern "C" {
#import "MiscTableViewPS.h"
#import <math.h>		// floor()
}

int const NUM_EDGES = 4;		// 4 edges to draw per focus rectangle
int const NUM_COORDS = 4;		// 4 coords per rectangle (x,y,w,h)
int const MAX_COORDS = NUM_EDGES * NUM_COORDS;	// 16 coords per 4 rects


//-----------------------------------------------------------------------------
// init_pswrap
//-----------------------------------------------------------------------------
static inline void init_pswrap()
	{
	static initialized = NO;
	if (!initialized)
		{
		initialized = YES;
		MISC_TV_initps();
		}
	}



@implementation MiscTableView(Cursor)
//=============================================================================
// CURSOR DRAWING
//=============================================================================
- (BOOL)isCursorEnabled { return (inhibitCursor == 0); }
- (void)disableCursor	{ inhibitCursor++; }
- (void)reenableCursor	{ inhibitCursor--; }


//-----------------------------------------------------------------------------
// - shouldDrawCursor
//-----------------------------------------------------------------------------
- (BOOL)shouldDrawCursor
	{
	Window* w = [self window];
	return ([self isCursorEnabled] && [self canDraw] && [w isKeyWindow] &&
				[w firstResponder] == self);
	}


//-----------------------------------------------------------------------------
// - getCursorSlot
//-----------------------------------------------------------------------------
- (MiscCoord_V)getCursorSlot
	{
	MiscTableBorder* const b = [self borderFor:tracker_border];
	MiscCoord_V vslot = b->getCursor();
	int const lim = b->count();
	if (lim > 0 && (vslot < 0 || vslot >= lim))
		{
		vslot = b->selectedSlot();
		if (vslot < 0)
			vslot = b->physicalToVisual([self firstVisibleSlot:tracker_border]);
		MiscAssert( 0 <= vslot );
		MiscAssert( vslot < lim );
		b->setCursor( vslot );
		}
	return vslot;
	}


//-----------------------------------------------------------------------------
// - getCursorFrame:
//-----------------------------------------------------------------------------
- (NXRect)getCursorFrame:(NXRect const*)clip
	{
	NXRect ret = {{ 0,0 }, { 0,0 }};
	if (row_border->count() > 0 && col_border->count() > 0)
		{
		MiscTableBorder* const b = [self borderFor:tracker_border];
		MiscCoord_P const pslot = b->visualToPhysical( [self getCursorSlot] );
		if (pslot >= 0)
			{
			NXRect r; [self getSlotInside:&r at:pslot from:tracker_border];
			if (NXIntersectsRect( &r, clip ))
				ret = r;
			}
		}
	return ret;
	}


//-----------------------------------------------------------------------------
// - getCursorEdges:clipTo:
//
//		Compute rectangles for drawing which make up the edges of the "focus
//		frame".	 Each edge is composed of values from the focus frame,
//		possibly including its: X, Y, MAXX, MAXY.
//-----------------------------------------------------------------------------
- (int)getCursorEdges:(NXRect*)edges clipTo:(NXRect const*)clip
	{
	struct TV_Lines { char xs, maxxs, xc, ys, maxys, yc, ws, wc, hs, hc; };
	static TV_Lines const LINES[ NUM_EDGES ] =
		{
			{ 1, 0,	 0, 1, 0,  0, 1,  0, 0,	 1 }, // top
			{ 1, 0,	 0, 0, 1, -1, 1,  0, 0,	 1 }, // bottom
			{ 1, 0,	 0, 1, 0,  1, 0,  1, 1, -2 }, // left
			{ 0, 1, -1, 1, 0,  1, 0,  1, 1, -2 }, // right
		};

	NXRect* edge = edges;
	NXRect r = [self getCursorFrame:clip];
	if (!NXEmptyRect( &r ))
		{
		for (int i = 0; i < NUM_EDGES; i++)
			{
			TV_Lines const& l = LINES[i];
			NXSetRect( edge, 
					l.xs * NX_X(&r) + l.maxxs * NX_MAXX(&r) + l.xc,
					l.ys * NX_Y(&r) + l.maxys * NX_MAXY(&r) + l.yc,
					l.ws * NX_WIDTH(&r) + l.wc,
					l.hs * NX_HEIGHT(&r) + l.hc );
			if (NXIntersectionRect( clip, edge ) != 0)
				edge++;
			}
		}
	return (edge - edges);
	}


//-----------------------------------------------------------------------------
// - getCursorCoords:clipTo:
//-----------------------------------------------------------------------------
- (int)getCursorCoords:(NXCoord*)coords clipTo:(NXRect const*)in_clip
	{
	int num_coords = 0;

	NXRect clip;
	if (in_clip == 0)
		[self getVisibleRect:&clip];
	else
		clip = *in_clip;

	NXRect edges[ NUM_EDGES ];
	int const num_edges = [self getCursorEdges:edges clipTo:&clip];
	if (num_edges > 0)
		{
		NXCoord* p = coords;
		for (int i = 0; i < num_edges; i++)
			{
			NXRect* edge = &edges[i];
			*p++ = NX_X( edge );
			*p++ = NX_Y( edge );
			*p++ = NX_WIDTH( edge );
			*p++ = NX_HEIGHT( edge );
			}
		num_coords = num_edges * NUM_COORDS;
		}
	return num_coords;
	}


//-----------------------------------------------------------------------------
// - drawCursorCoords:count:
//-----------------------------------------------------------------------------
- (void)drawCursorCoords:(NXCoord const*)coords count:(int)n
	{
	init_pswrap();
	BOOL needsFocus = ![self isFocusView];
	if (needsFocus) [self lockFocus];
	MISC_TV_dashedrects( coords, n );
	if (needsFocus) [self unlockFocus];
	}


//-----------------------------------------------------------------------------
// - drawCursorClipTo:
//-----------------------------------------------------------------------------
- (void)drawCursorClipTo:(NXRect const*)clip
	{
	if ([self isCursorEnabled])
		{
		NXCoord coords[ MAX_COORDS ];
		int const num_coords = [self getCursorCoords:coords clipTo:clip];
		if (num_coords > 0)
			{
			[self drawCursorCoords:coords count:num_coords];
			cursorSlot = [self borderFor:tracker_border]->getCursor();
			}
		}
	}


//-----------------------------------------------------------------------------
// - drawCursor
//-----------------------------------------------------------------------------
- (void)drawCursor
	{
	[self drawCursorClipTo:0];
	}


//-----------------------------------------------------------------------------
// - eraseCursor
//-----------------------------------------------------------------------------
- (void)eraseCursor
	{
	if ([self isCursorEnabled])
		{
		MiscTableBorder* const b = [self borderFor:tracker_border];
		if (cursorSlot >= 0 && cursorSlot < b->count())
			{
			[self disableCursor];

			MiscCoord_P s = b->visualToPhysical( cursorSlot );
			if (tracker_border == MISC_COL_BORDER)
				[self drawCol:s];
			else
				[self drawRow:s];

			[self reenableCursor];
			cursorSlot = -1;
			}
		}
	}


//-----------------------------------------------------------------------------
// - becomeFirstResponder
//-----------------------------------------------------------------------------
- (id)becomeFirstResponder
	{
	Window* win = [self window];
	if (win && [win isKeyWindow])
		{
		if ([self shouldDrawCursor])
			{
			[self drawCursor];
			[win flushWindow];
			}
		[[FontManager new] setSelFont:[scroll font] isMultiple:NO];
		}
	return self;
	}


//-----------------------------------------------------------------------------
// - resignFirstResponder
//-----------------------------------------------------------------------------
- (id)resignFirstResponder
	{
	[self eraseCursor];
	[[self window] flushWindow];
	return self;
	}


//-----------------------------------------------------------------------------
// - becomeKeyWindow
//-----------------------------------------------------------------------------
- (id)becomeKeyWindow
	{
	if ([self shouldDrawCursor])
		{
		[self drawCursor];
		[[self window] flushWindow];
		}
	[[FontManager new] setSelFont:[scroll font] isMultiple:NO];
	return self;
	}


//-----------------------------------------------------------------------------
// - resignKeyWindow
//-----------------------------------------------------------------------------
- (id)resignKeyWindow
	{
	[self eraseCursor];
	[[self window] flushWindow];
	return self;
	}

//-----------------------------------------------------------------------------
// - reflectCursor
//-----------------------------------------------------------------------------
- (void)reflectCursor
	{
	if ([self shouldDrawCursor])
		{
		Window* w = [self window];
		[w disableFlushWindow];
		[self eraseCursor];
		[self drawCursor];
		[w reenableFlushWindow];
		[w flushWindow];
		}
	}

//-----------------------------------------------------------------------------
// - moveCursorBy:
//-----------------------------------------------------------------------------
- (void)moveCursorBy:(int)delta
	{
	MiscTableBorder* const b = [self borderFor:tracker_border];
	int const lim = b->count();
	if (lim > 0)
		{
		MiscCoord_V slot = b->getCursor() + delta;
		if (slot < 0)
			slot = lim - 1;
		else if (slot >= lim)
			slot = 0;
		b->setCursor( slot );

		Window* const w = [self window];
		[w disableFlushWindow];
		[self reflectCursor];
		[self border:tracker_border scrollToVisible:b->visualToPhysical(slot)];
		[w reenableFlushWindow];
		[w flushWindow];
		}
	}


//-----------------------------------------------------------------------------
// - keyboardSelect:
//-----------------------------------------------------------------------------
- (void)keyboardSelect:(NXEvent const*)p
	{
	[tracker mouseDown:p atPos:[self borderFor:tracker_border]->getCursor()];
	}

@end
