//=============================================================================
//
//	Copyright (C) 1995,1996,1997,1998 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//
//=============================================================================
//-----------------------------------------------------------------------------
// MiscTableViewDrag.M
//
//		Image dragging methods for MiscTableView.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscTableViewDrag.M,v 1.2 98/03/24 04:46:33 sunshine Exp $
// $Log:		MiscTableViewDrag.M,v $
//	Revision 1.2  98/03/24	04:46:33  sunshine
//	v138: Implemented -shouldDelayWindowOrderingForEvent:
//	
//	Revision 1.1  97/11/14	20:43:54  zarnuk
//	v130: Image dragging methods.
//-----------------------------------------------------------------------------
#import "MiscTableViewPrivate.h"
#import "MiscTableScrollPrivate.h"
#import <misckit/MiscTableCell.h>
extern "Objective-C" {
#import <appkit/Application.h>
}
extern "C" {
#import <math.h>		// floor()
}

typedef MiscDelegateFlags DF;

@implementation MiscTableView(Drag)

//=============================================================================
// IMAGE DRAGGING
//=============================================================================
static inline float absval( float x ) { return (x < 0 ? -x : x ); }
static inline BOOL isSlop( NXEvent* e1, NXEvent* e2, float const slop )
	{
	return (absval(e1->location.x - e2->location.x) <= slop &&
			absval(e1->location.y - e2->location.y) <= slop);
	}


//-----------------------------------------------------------------------------
// draggingSourceOperationMaskForLocal:
//-----------------------------------------------------------------------------
- (NXDragOperation)draggingSourceOperationMaskForLocal:(BOOL)flag
	{
	id const d = [scroll responsibleDelegate:DF::DEL_DRAG_OP_MASK];
	if (d != 0)
		return [d tableScroll:scroll draggingSourceOperationMaskForLocal:flag];
	return NX_DragOperationGeneric;
	}


//-----------------------------------------------------------------------------
// ignoreModifierKeysWhileDragging
//-----------------------------------------------------------------------------
- (BOOL)ignoreModifierKeysWhileDragging
	{
	id const d = [scroll responsibleDelegate:DF::DEL_DRAG_IGNORE_MODIFIERS];
	if (d != 0)
		return [d tableScrollIgnoreModifierKeysWhileDragging:scroll];
	return YES;
	}


//-----------------------------------------------------------------------------
// shouldDelayWindowOrderingForEvent:
//-----------------------------------------------------------------------------
- (BOOL)shouldDelayWindowOrderingForEvent:(NXEvent*)e
	{
	BOOL delay = NO;
	NXPoint p = e->location;
	[self convertPoint:&p fromView:0];
	MiscCoord_P r,c;
	if ([self getRow:&r andCol:&c forPoint:&p] &&
		[self canPerformDrag:e at:r:c])
		{
		id d = [scroll responsibleDelegate:DF::DEL_DRAG_DELAY_WIN_ORDERING];
		if (d != 0)
			delay = [d tableScroll:scroll shouldDelayWindowOrderingForEvent:e];
		else
			delay = YES;
		}
	return delay;
	}


//-----------------------------------------------------------------------------
// getDragImageAt::
//-----------------------------------------------------------------------------
- (NXImage*)getDragImageAt:(MiscCoord_P)row :(MiscCoord_P)col
	{
	NXImage* i = 0;
	id const d = [scroll responsibleDelegate:DF::DEL_IMAGE_FOR_DRAG];
	if (d != 0)
		i = [d tableScroll:scroll imageForDragOperationAt:row:col];
	if (i == 0)
		{
		id cell = [scroll cellAt:row:col];
		if ([cell type] == NX_ICONCELL)
			i = [cell image];
		}
	return i;
	}


//-----------------------------------------------------------------------------
// calcOrigin:andOffset:forImage:inRect:at::downEvent:dragEvent:
//
// Compute the origin of the image and the offset of the mouse for dragging 
// via View's -dragImage:at:offset:... method.	There are two cases: 
//
// CASE *1*: The image being dragged fits exactly in the Cell's image area.	 
//	  In this case drag from the point where the mouse went down inside the 
//	  image's boundry.	This give the visual impression of lifting the image 
//	  exactly from where it is sitting; without it jumping to some other 
//	  location.	 
// CASE *2*: The image is a different size than the Cell's image area.	In 
//	  this case just center the image under the mouse.	
//-----------------------------------------------------------------------------
- (void)calcOrigin:(NXPoint*)origin
	andOffset:(NXPoint*)offset
	forImage:(NXImage*)image
	inRect:(NXRect const*)rect
	at:(MiscCoord_P)row :(MiscCoord_P)col
	downEvent:(NXEvent*)downEvent
	dragEvent:(NXEvent*)dragEvent
	{
	offset->x = 0;
	offset->y = 0;

	NXSize imageSize;
	[image getSize:&imageSize];

	id cell = [scroll cellAt:row:col];
	NXRect iconRect = *rect;
	[cell getIconRect:&iconRect];

	if (imageSize.width	 == iconRect.size.width &&				// CASE *1*
		imageSize.height == iconRect.size.height)
		{
		*origin = iconRect.origin;
		if ([self isFlipped])
			origin->y += iconRect.size.height;

		if (dragEvent != 0 && dragEvent->type == NX_MOUSEDRAGGED)
			{
			offset->x = dragEvent->location.x - downEvent->location.x;
			offset->y = dragEvent->location.y - downEvent->location.y;
			}
		}
	else														// CASE *2*
		{
		*origin = downEvent->location;
		[self convertPoint:origin fromView:nil];
		origin->x -= floor( imageSize.width / 2 );
		origin->y -= floor( imageSize.height / 2 ) *
								([self isFlipped] ? -1 : 1);
		}
	}


//-----------------------------------------------------------------------------
// prepareDragPasteboard:at::
//-----------------------------------------------------------------------------
- (void)prepareDragPasteboard:(Pasteboard*)pb
	at:(MiscCoord_P)row :(MiscCoord_P)col
	{
	id const d = [scroll responsibleDelegate:DF::DEL_PREPARE_PB_FOR_DRAG];
	if (d != 0)
		[d tableScroll:scroll preparePasteboard:pb forDragOperationAt:row:col];
	}


//-----------------------------------------------------------------------------
// performDrag:at::inRect:dragEvent:
//-----------------------------------------------------------------------------
- (BOOL)performDrag:(NXEvent*)mouseDown
	at:(MiscCoord_P)row :(MiscCoord_P)col
	inRect:(NXRect const*)r
	dragEvent:(NXEvent*)dragEvent
	{
	BOOL ret = NO;

	NXImage* i = [self getDragImageAt:row:col];
	if (i != 0)
		{
		NXPoint origin, offset;
		[self calcOrigin:&origin andOffset:&offset forImage:i inRect:r
				at:row:col downEvent:mouseDown dragEvent:dragEvent];
	
		Pasteboard* pb = [Pasteboard newName:NXDragPboard];
		[self prepareDragPasteboard:pb at:row:col];
	
		[self dragImage:i at:&origin offset:&offset
				event:mouseDown pasteboard:pb source:self slideBack:YES];
		ret = YES;
		}
	return ret;
	}


//-----------------------------------------------------------------------------
// awaitDragEvent:at::
//-----------------------------------------------------------------------------
- (BOOL)awaitDragEvent:(NXEvent*)p
	at:(MiscCoord_P)row :(MiscCoord_P)col
	inRect:(NXRect const*)rect
	{
	float const DELAY = 0.25;
	float const SLOP = 4.0;
	int const WANTED = (NX_MOUSEUPMASK | NX_MOUSEDRAGGEDMASK);

	BOOL ret = NO;
	NXEvent mouseDown = *p;
	NXEvent* event;
	NXEvent peeker;
	int old_mask = [[self window] addToEventMask:WANTED];

	do	{
		event = [NXApp peekNextEvent:WANTED into:&peeker waitFor:DELAY
				threshold:NX_MODALRESPTHRESHOLD];
		if (event != 0 && event->type == NX_MOUSEDRAGGED)
			event = [NXApp getNextEvent:NX_MOUSEDRAGGEDMASK];
		} while (event != 0 && event->type == NX_MOUSEDRAGGED && 
				isSlop( event, &mouseDown, SLOP ));

	[[self window] setEventMask:old_mask];

	if (event == 0 || event->type == NX_MOUSEDRAGGED)
		ret = [self performDrag:&mouseDown at:row:col inRect:rect
						dragEvent:event];
	return ret;
	}


//-----------------------------------------------------------------------------
// canPerformDrag:at::
//-----------------------------------------------------------------------------
- (BOOL)canPerformDrag:(NXEvent*)p at:(MiscCoord_P)r :(MiscCoord_P)c
	{
	id const d = [scroll responsibleDelegate:DF::DEL_ALLOW_DRAG];
	return (d != 0 && [d tableScroll:scroll allowDragOperationAt:r:c] &&
			[scroll responsibleDelegate:DF::DEL_PREPARE_PB_FOR_DRAG] != 0 &&
		   ([[scroll cellAt:r:c] type] == NX_ICONCELL ||
			[scroll responsibleDelegate:DF::DEL_IMAGE_FOR_DRAG] != 0));
	}

@end
